/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.kafka.connect.runtime.rest.prometheus;

import io.prometheus.client.Gauge;
import org.apache.kafka.common.MetricName;
import org.apache.kafka.common.MetricNameTemplate;
import org.apache.kafka.common.metrics.JmxReporter;
import org.apache.kafka.common.metrics.KafkaMetric;
import org.apache.kafka.common.metrics.MetricConfig;
import org.apache.kafka.common.utils.MockTime;
import org.apache.kafka.common.utils.Time;
import org.apache.kafka.connect.runtime.ConnectMetricsRegistry;
import org.junit.Test;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static org.junit.Assert.assertEquals;

import static org.apache.kafka.connect.runtime.ConnectMetrics.JMX_PREFIX;

public class PrometheusConnectMetricsServletTest {

    private Time time = new MockTime();
    private ConnectMetricsRegistry registry = new ConnectMetricsRegistry();
    private long metricValue = 49;

    private Set<MetricNameTemplate> connectMetricNames = new HashSet<>(Arrays.asList(registry.connectorCount,
        registry.connectorStartupAttemptsTotal, registry.connectorStartupFailureTotal, registry.connectorStartupFailureTotal,
        registry.taskCount, registry.taskStartupAttemptsTotal, registry.taskStartupFailureTotal, registry.taskStartupSuccessTotal,
        registry.rebalanceCompletedTotal, registry.rebalanceTimeAvg, registry.rebalanceTimeMax, registry.epoch, registry.rebalanceMode,
        registry.rebalanceTimeSinceLast));
    private Set<MetricNameTemplate> taskMetricNames = new HashSet<>(Arrays.asList(registry.taskCommitTimeAvg, registry.taskCommitTimeMax,
        registry.taskCommitSuccessPercentage, registry.taskCommitFailurePercentage, registry.taskRunningRatio, registry.taskPauseRatio,
        registry.taskBatchSizeAvg, registry.taskBatchSizeMax, registry.sourceRecordWriteRate, registry.sourceRecordPollBatchTimeAvg,
        registry.sourceRecordPollBatchTimeMax, registry.sourceRecordWriteTotal, registry.sourceRecordWriteTotal, registry.sourceRecordPollTotal,
        registry.sourceRecordPollRate, registry.sourceRecordActiveCount, registry.sourceRecordActiveCountMax, registry.sourceRecordActiveCountAvg,
        registry.sinkRecordPartitionCount, registry.sinkRecordLagMax, registry.sinkRecordSendRate, registry.sinkRecordPutBatchTimeAvg,
        registry.lastErrorTimestamp, registry.recordProcessingFailures, registry.recordProcessingErrors, registry.recordsSkipped,
        registry.retries, registry.errorsLogged));

    private Set<String> expectedPrometheusConnectMetrics = new HashSet<>(Arrays.asList(
        "kafka_connect_worker_connector_startup_failure_total", "kafka_connect_worker_rebalance_rebalance_max_time_ms", "kafka_connect_worker_rebalance_epoch",
        "kafka_connect_worker_rebalance_rebalance_avg_time_ms", "kafka_connect_worker_connector_count", "kafka_connect_worker_rebalance_rebalancing",
        "kafka_connect_worker_connector_startup_attempts_total", "kafka_connect_worker_rebalance_time_since_last_rebalance_ms",
        "kafka_connect_worker_task_startup_failure_total", "kafka_connect_worker_task_count", "kafka_connect_worker_task_startup_attempts_total",
        "kafka_connect_worker_task_startup_success_total", "kafka_connect_worker_rebalance_completed_rebalances_total"));
    private Set<String> expectedPrometheusTaskMetrics = new HashSet<>(Arrays.asList(
        "kafka_connect_connector_task_offset_commit_avg_time_ms", "kafka_connect_connector_task_batch_size_max", "kafka_connect_sink_task_partition_count",
        "kafka_connect_connector_task_running_ratio", "kafka_connect_source_task_source_record_active_count_avg", "kafka_connect_source_task_source_record_poll_rate",
        "kafka_connect_task_error_total_record_errors", "kafka_connect_source_task_poll_batch_max_time_ms", "kafka_connect_source_task_source_record_active_count_max",
        "kafka_connect_sink_task_sink_record_lag_max", "kafka_connect_task_error_total_errors_logged", "kafka_connect_connector_task_offset_commit_failure_percentage",
        "kafka_connect_source_task_source_record_poll_total", "kafka_connect_sink_task_sink_record_send_rate", "kafka_connect_source_task_source_record_write_total",
        "kafka_connect_source_task_source_record_active_count", "kafka_connect_task_error_total_record_failures", "kafka_connect_task_error_total_records_skipped",
        "kafka_connect_task_error_total_retries", "kafka_connect_source_task_poll_batch_avg_time_ms", "kafka_connect_sink_task_put_batch_avg_time_ms",
        "kafka_connect_task_error_last_error_timestamp", "kafka_connect_connector_task_offset_commit_success_percentage", "kafka_connect_connector_task_batch_size_avg",
        "kafka_connect_connector_task_offset_commit_max_time_ms", "kafka_connect_connector_task_pause_ratio", "kafka_connect_source_task_source_record_write_rate"));

    @Test
    public void testPrometheusConnectMetricsServlet() {
        List<KafkaMetric> metricsList = new ArrayList<>();
        String connectName = "TestConnector";
        String task = "2";
        Map<String, String> tags = new LinkedHashMap<>();
        tags.put("connector", connectName);
        tags.put("task", task);

        for (MetricNameTemplate template : connectMetricNames) {
            metricsList.add(createKafkaMetric(template, Collections.emptyMap()));
        }

        for (MetricNameTemplate template : taskMetricNames) {
            metricsList.add(createKafkaMetric(template, tags));
        }

        JmxReporter jmxReporter = new JmxReporter(JMX_PREFIX);
        jmxReporter.init(metricsList);

        PrometheusConnectMetricsServlet servlet = new PrometheusConnectMetricsServlet();
        servlet.createPrometheusMetrics();

        Map<String, Gauge> gauges = servlet.getAllGauges();

        assertEquals(40, gauges.size());

        for (String prometheusName : expectedPrometheusConnectMetrics) {
            assertEquals(metricValue, (long) gauges.get(prometheusName).get());
        }

        for (String prometheusName : expectedPrometheusTaskMetrics) {
            assertEquals(metricValue, (long) gauges.get(prometheusName).labels(connectName, task).get());
        }
    }

    private MetricName createMetricName(MetricNameTemplate template, Map<String, String> tags) {
        return new MetricName(template.name(), template.group(), template.description(), tags);
    }

    private KafkaMetric createKafkaMetric(MetricNameTemplate template, Map<String, String> tags) {
        return new KafkaMetric(new Object(), createMetricName(template, tags), new ImmutableValue<>(metricValue), null, time);
    }

    static class ImmutableValue<T> implements org.apache.kafka.common.metrics.Gauge<T> {
        private final T value;

        public ImmutableValue(T value) {
            this.value = value;
        }

        @Override
        public T value(MetricConfig config, long now) {
            return value;
        }
    }

}
