/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package prometheus.metrics.reporting

import java.util.concurrent.TimeUnit

import com.cloudera.kafka.prometheus.metrics.reporting.JmxMetricNames._
import com.cloudera.kafka.prometheus.metrics.reporting.{GroupMetaData, PrometheusMetricsHandler, PrometheusMetricsServlet}
import com.yammer.metrics.Metrics
import com.yammer.metrics.core.MetricsRegistry
import io.prometheus.client.CollectorRegistry
import kafka.common.OffsetAndMetadata
import kafka.coordinator.group.{GroupSummary, MemberSummary}
import kafka.metrics.KafkaMetricsGroup
import kafka.network.RequestMetrics.TotalTimeMs
import org.apache.kafka.clients.consumer.ConsumerPartitionAssignor.Assignment
import org.apache.kafka.clients.consumer.internals.ConsumerProtocol
import org.apache.kafka.common.TopicPartition
import org.apache.kafka.common.utils.Time
import org.junit.Assert.assertEquals
import org.junit.{After, Before, Test}
import org.mockito.Mockito
import scala.collection.JavaConverters._
import scala.collection.immutable


class PrometheusMetricsTest extends KafkaMetricsGroup {

  val prometheusMetricsServlet: PrometheusMetricsServlet = Mockito.spy(new PrometheusMetricsServlet(null))
  val prometheusMetricsHandler: PrometheusMetricsHandler = Mockito.spy(prometheusMetricsServlet.getPrometheusMetricsHandler)
  val metricsRegistry: MetricsRegistry = Metrics.defaultRegistry()

  @Before
  def setup(): Unit = {
    Metrics.defaultRegistry().addListener(prometheusMetricsServlet)
    Mockito.reset(prometheusMetricsServlet)
    prometheusMetricsServlet.clearPreviousMetrics()
  }

  @After
  def cleanup(): Unit = {
    Metrics.defaultRegistry().removeListener(prometheusMetricsServlet)
    Metrics.defaultRegistry().allMetrics().forEach({
      case (metricName, _) => Metrics.defaultRegistry().removeMetric(metricName)
    })
  }

  @Test
  def testPrometheusMetricsServlet(): Unit = {
    val topic = "testTopic"
    val streamTopic = "streamTopic"
    val partition = "1"
    val partition2 = "2"
    val clientId1 = "testClientId1"
    val clientId2 = "testClientId2"
    val clientHost = "testClientHost"
    val groupId = "testGroupId"
    val groupState = "Stable"
    val tp1 = new TopicPartition(topic, partition.toInt)
    val tp2 = new TopicPartition(topic, partition2.toInt)
    val tp3 = new TopicPartition(streamTopic, partition.toInt)
    val assignment1 = ConsumerProtocol.serializeAssignment(new Assignment(List(tp1).asJava))
    val assignment2 = ConsumerProtocol.serializeAssignment(new Assignment(List(tp2).asJava))
    val assignment3 = ConsumerProtocol.serializeAssignment(new Assignment(List(tp3).asJava))
    val assignment1ByteArray = new Array[Byte](assignment1.remaining())
    val assignment2ByteArray = new Array[Byte](assignment2.remaining())
    val assignment3ByteArray = new Array[Byte](assignment3.remaining())
    assignment1.get(assignment1ByteArray)
    assignment2.get(assignment2ByteArray)
    assignment3.get(assignment3ByteArray)

    val memberSummary1 = MemberSummary("memberId1", Some("testInstanceid1"), clientId1, clientHost, new Array[Byte](10), assignment1ByteArray)
    val memberSummary2 = MemberSummary("memberId2", Some("testInstanceid2"), clientId2, clientHost, new Array[Byte](10), assignment2ByteArray)
    val memberSummary3 = MemberSummary("memberId3", Some("testInstanceid3"), clientId2, clientHost, new Array[Byte](10), assignment3ByteArray)
    val groupSummary = GroupSummary(groupState, "consumer", "roundrobin", List(memberSummary1, memberSummary2, memberSummary3))
    val logEndOffset = 521L
    val committedOffset1 = 492
    val committedOffset2 = 599
    val committedTimestamp = Time.SYSTEM.milliseconds()
    val offsets = immutable.Map(tp1 -> OffsetAndMetadata(committedOffset1, "", committedTimestamp),
      tp2 -> OffsetAndMetadata(committedOffset2, "", committedTimestamp))
    val groupMetaData = GroupMetaData(groupId, groupSummary, offsets)

    setupConsumerRelatedMock(topic, partition.toInt, List(groupMetaData))

    // Broker level metrics
    val brokerBytesInPerSec = 508
    val brokerBytesOutPerSec = 411
    val brokerUncleanLeaderPerSec = 12
    val brokerMessagesInPerSec = 602
    val partitionCount = 33
    val leaderCount = 11
    val underReplicatedPartitions = 3
    val networkProcessorAvgIdlePercent = 51.0
    val activeControllerCount = 1
    val offlinePartitionsCount = 6

    metricsRegistry.newMeter(explicitMetricName(KafkaServerGroup, BrokerTopicMetrics, BytesInPerSec, Map.empty), "bytes",  TimeUnit.SECONDS)
      .mark(brokerBytesInPerSec)
    metricsRegistry.newMeter(explicitMetricName(KafkaServerGroup, BrokerTopicMetrics, BytesOutPerSec, Map.empty), "bytes",  TimeUnit.SECONDS)
      .mark(brokerBytesOutPerSec)
    metricsRegistry.newMeter(explicitMetricName(KafkaControllerGroup, ControllerStats, UncleanLeaderElectionsPerSec, Map.empty), "elections",  TimeUnit.SECONDS)
      .mark(brokerUncleanLeaderPerSec)
    metricsRegistry.newMeter(explicitMetricName(KafkaServerGroup, BrokerTopicMetrics, MessagesInPerSec, Map.empty), "elections",  TimeUnit.SECONDS)
      .mark(brokerMessagesInPerSec)
    metricsRegistry.newGauge(explicitMetricName(KafkaServerGroup, ReplicaManager, PartitionCount, Map.empty), () => partitionCount)
    metricsRegistry.newGauge(explicitMetricName(KafkaServerGroup, ReplicaManager, LeaderCount, Map.empty), () => leaderCount)
    metricsRegistry.newGauge(explicitMetricName(KafkaServerGroup, ReplicaManager, UnderReplicatedPartitions, Map.empty), () => underReplicatedPartitions)
    metricsRegistry.newGauge(explicitMetricName(KafkaNetworkGroup, SocketServer, NetworkProcessorAvgIdlePercent, Map.empty), () => networkProcessorAvgIdlePercent)
    metricsRegistry.newGauge(explicitMetricName(KafkaControllerGroup, KafkaController, ActiveControllerCount, Map.empty), () => activeControllerCount)
    metricsRegistry.newGauge(explicitMetricName(KafkaControllerGroup, KafkaController, OfflinePartitionsCount, Map.empty), () => offlinePartitionsCount)

    val produceTotalTimeMs = 80273
    val requestProduceTag = Map("request" -> "Produce")
    metricsRegistry.newHistogram(explicitMetricName(KafkaNetworkGroup, RequestMetrics, TotalTimeMs, requestProduceTag), true)
      .update(produceTotalTimeMs)

    val fetchConsumerTotalTimeMs = 5039
    val requestFetchConsumerTag = Map("request" -> "FetchConsumer")
    metricsRegistry.newHistogram(explicitMetricName(KafkaNetworkGroup, RequestMetrics, TotalTimeMs, requestFetchConsumerTag), true)
      .update(fetchConsumerTotalTimeMs)

    // Topic level metrics
    val topicMessagesInPerSec = 225
    val topicBytesInPerSec = 351
    val topicBytesOutPerSec = 672

    val topicLevelTag = Map("topic" -> topic)
    metricsRegistry.newMeter(explicitMetricName(KafkaServerGroup, BrokerTopicMetrics, MessagesInPerSec, topicLevelTag), "messages",  TimeUnit.SECONDS)
      .mark(topicMessagesInPerSec)
    metricsRegistry.newMeter(explicitMetricName(KafkaServerGroup, BrokerTopicMetrics, BytesInPerSec, topicLevelTag), "bytes",  TimeUnit.SECONDS)
      .mark(topicBytesInPerSec)
    metricsRegistry.newMeter(explicitMetricName(KafkaServerGroup, BrokerTopicMetrics, BytesOutPerSec, topicLevelTag), "bytes",  TimeUnit.SECONDS)
      .mark(topicBytesOutPerSec)

    // Partition level metrics
    val partitionMessagesInPerSec = 211
    val partitionBytesInPerSec = 98
    val partitionBytesOutPerSec = 27
    val replicasCount = 22
    val inSyncReplicasCount = 39
    val underReplicated = 4

    val partitionLevelTags = Map("topic" -> topic, "partition" -> partition)
    metricsRegistry.newMeter(explicitMetricName(KafkaServerGroup, BrokerTopicMetrics, MessagesInPerSec, partitionLevelTags), "messages",  TimeUnit.SECONDS)
      .mark(partitionMessagesInPerSec)
    metricsRegistry.newMeter(explicitMetricName(KafkaServerGroup, BrokerTopicMetrics, BytesInPerSec, partitionLevelTags), "bytes",  TimeUnit.SECONDS)
      .mark(partitionBytesInPerSec)
    metricsRegistry.newMeter(explicitMetricName(KafkaServerGroup, BrokerTopicMetrics, BytesOutPerSec, partitionLevelTags), "bytes",  TimeUnit.SECONDS)
      .mark(partitionBytesOutPerSec)
    metricsRegistry.newGauge(explicitMetricName(KafkaClusterGroup, PartitionType, ReplicasCount, partitionLevelTags), () => replicasCount)
    metricsRegistry.newGauge(explicitMetricName(KafkaClusterGroup, PartitionType, InSyncReplicasCount, partitionLevelTags), () => inSyncReplicasCount)
    metricsRegistry.newGauge(explicitMetricName(KafkaClusterGroup, PartitionType, UnderReplicated, partitionLevelTags), () => underReplicated)
    metricsRegistry.newGauge(explicitMetricName(KafkaLog, Log, LogEndOffset, partitionLevelTags), () => logEndOffset)

    // Not allowed metrics. These should be ignored by PrometheusMetricsServlet
    metricsRegistry.newGauge(explicitMetricName(KafkaLog, Log, "nonExistentMetric", partitionLevelTags), () => 1)
    metricsRegistry.newGauge(explicitMetricName(KafkaClusterGroup, PartitionType, "nonExistentMetric", partitionLevelTags), () => 1)
    metricsRegistry.newGauge(explicitMetricName(KafkaNetworkGroup, SocketServer, "nonExistentMetric", Map.empty), () => 1)
    metricsRegistry.newGauge(explicitMetricName(KafkaControllerGroup, ControllerStats, "nonExistentMetric", Map.empty), () => 1)
    metricsRegistry.newGauge(explicitMetricName(KafkaServerGroup, BrokerTopicMetrics, "nonExistentMetric", Map.empty), () => 1)
    metricsRegistry.newGauge(explicitMetricName("nonExistentMetric", BrokerTopicMetrics, "nonExistentMetric", Map.empty), () => 1)

    // Client Id level metrics
    val clientMessagesInPerSec = 89

    val clientLevelTags = Map("clientId"-> clientId1, "topic" -> topic, "partition" -> partition)
    metricsRegistry.newMeter(explicitMetricName(KafkaServerGroup, BrokerClientMetrics, MessagesInPerSec, clientLevelTags), "messages",  TimeUnit.SECONDS)
      .mark(clientMessagesInPerSec)

    prometheusMetricsServlet.updatePrometheusMetrics(false)
    val gauges = prometheusMetricsServlet.getPrometheusGauges

    // Verify all 40 metrics are registered by Prometheus
    verifyAllMetricsAreRegistered()

    def checkAllMetrics(): Unit = {
      // Check Consumer Metrics
      assertEquals(gauges("partition_log_endoffset").labels(topic, partition).get().toLong, logEndOffset)
      assertEquals(gauges("group_committed_offset").labels(groupId, topic, partition).get().toLong, committedOffset1)
      assertEquals(gauges("group_committed_offset").labels(groupId, topic, partition2).get().toLong, committedOffset2)

      // Check Kafka Increased Count Metrics
      assertEquals(gauges("broker_bytesinpersec_total").get().toLong, brokerBytesInPerSec)
      assertEquals(gauges("broker_bytesoutpersec_total").get().toLong, brokerBytesOutPerSec)
      assertEquals(gauges("broker_messagesinpersec_total").get().toLong, brokerMessagesInPerSec)
      assertEquals(gauges("broker_uncleanleaderelectionspersec_total").get().toLong, brokerUncleanLeaderPerSec)
      assertEquals(gauges("topic_bytesinpersec_total").labels(topic).get().toLong, topicBytesInPerSec)
      assertEquals(gauges("topic_bytesoutpersec_total").labels(topic).get().toLong, topicBytesOutPerSec)
      assertEquals(gauges("topic_messagesinpersec_total").labels(topic).get().toLong, topicMessagesInPerSec)
      assertEquals(gauges("topic_partition_messagesinpersec_total").labels(topic, partition).get().toLong, partitionMessagesInPerSec)
      assertEquals(gauges("topic_partition_bytesinpersec_total").labels(topic, partition).get().toLong, partitionBytesInPerSec)
      assertEquals(gauges("topic_partition_bytesoutpersec_total").labels(topic, partition).get().toLong, partitionBytesOutPerSec)
      assertEquals(gauges("broker_producer_messagesinpersec_total").labels(clientId1, topic, partition).get().toLong, clientMessagesInPerSec)

      // Check Other Kafka Metrics
      assertEquals(gauges("broker_partitioncount").get().toInt, partitionCount)
      assertEquals(gauges("broker_leadercount").get().toInt, leaderCount)
      assertEquals(gauges("broker_underreplicatedpartitions").get().toInt, underReplicatedPartitions)
      assertEquals(gauges("broker_networkprocessoravgidlepercent").get().toInt, networkProcessorAvgIdlePercent.toInt)
      assertEquals(gauges("broker_activecontrollercount").get().toInt, activeControllerCount)
      assertEquals(gauges("broker_offlinepartitionscount").get().toInt, offlinePartitionsCount)
      assertEquals(gauges("topic_partition_replicascount").labels(topic, partition).get().toInt, replicasCount)
      assertEquals(gauges("topic_partition_insyncreplicascount").labels(topic, partition).get().toInt, inSyncReplicasCount)
      assertEquals(gauges("topic_partition_underreplicated").labels(topic, partition).get().toInt, underReplicated)
      assertEquals(gauges("broker_totaltimems_produce_99thpercentile").get().toInt, produceTotalTimeMs)
      assertEquals(gauges("broker_totaltimems_fetchconsumer_99thpercentile").get().toInt, fetchConsumerTotalTimeMs)
    }

    // Check Consumer Metrics, Kafka Increased Count Metrics, Other Kafka Metrics are marked correctly
    checkAllMetrics()

    // Updating the Prometheus metrics should provide the same marked result as before since the previous call didn't update the cache (false flag)
    prometheusMetricsServlet.updatePrometheusMetrics(true)
    checkAllMetrics()

    // Kafka Increased Count Metrics aren't marked between 2 calls, therefore the increase count should be 0 for all
    prometheusMetricsServlet.updatePrometheusMetrics(true)
    assertEquals(gauges("broker_bytesinpersec_total").get().toLong, 0)
    assertEquals(gauges("broker_bytesoutpersec_total").get().toLong, 0)
    assertEquals(gauges("broker_messagesinpersec_total").get().toLong, 0)
    assertEquals(gauges("broker_uncleanleaderelectionspersec_total").get().toLong, 0)
    assertEquals(gauges("topic_bytesinpersec_total").labels(topic).get().toLong, 0)
    assertEquals(gauges("topic_bytesoutpersec_total").labels(topic).get().toLong, 0)
    assertEquals(gauges("topic_messagesinpersec_total").labels(topic).get().toLong, 0)
    assertEquals(gauges("topic_partition_messagesinpersec_total").labels(topic, partition).get().toLong, 0)
    assertEquals(gauges("topic_partition_bytesinpersec_total").labels(topic, partition).get().toLong, 0)
    assertEquals(gauges("topic_partition_bytesoutpersec_total").labels(topic, partition).get().toLong, 0)
    assertEquals(gauges("broker_producer_messagesinpersec_total").labels(clientId1, topic, partition).get().toLong, 0)

    // Remove metrics and check consumer metrics
    prometheusMetricsServlet.clearPreviousMetrics()
    assertEquals(gauges("partition_log_endoffset").labels(topic, partition).get().toLong, 0)
    assertEquals(gauges("group_committed_offset").labels(groupId, topic, partition).get().toLong, 0)
    assertEquals(gauges("group_committed_offset").labels(groupId, topic, partition2).get().toLong, 0)
  }

  private def verifyAllMetricsAreRegistered():Unit = {
    val registeredPrometheusMetrics = CollectorRegistry.defaultRegistry.metricFamilySamples()
    var counter = 0

    while (registeredPrometheusMetrics.hasMoreElements) {
      registeredPrometheusMetrics.nextElement()
      counter += 1
    }

    assertEquals(39, counter)
  }

  private def setupConsumerRelatedMock(topic: String, partition: Int, groupMetaDatas: List[GroupMetaData]) = {
    Mockito.doReturn(groupMetaDatas, Nil: _*).when(prometheusMetricsHandler).getValidConsumerGroups
    Mockito.doReturn(prometheusMetricsHandler, Nil: _*).when(prometheusMetricsServlet).getPrometheusMetricsHandler
  }

}